# PrettyCode.py
#
# This code is used to simplify large transfer function coefficients so that they may be computed efficiently.
# There are also functions that mimic MATLAB functionality
#
# Chet Gnegy
# chetgnegy@gmail.com

from timeit import default_timer as timer # Remove

from sympy import *
from operator import itemgetter
import copy

from sympy.core import factor_terms


# Extracts the coefficients of a function for a variable var
# @params fun     The expression that we are extracting coefficients from
# @params var     The variable of interest
# returns         A tuple (c,v), where c is a list of coefficients for decending powers 
#                 of var, and v is the respective powers of var
def collect_terms(fun, var):
  coef = Poly(fun, var).all_coeffs()
  var_array = [var**(len(coef)-i-1) for i in range(0,len(coef))]
  return ( coef, var_array)

# A wrapper for equality checking for symbols
def equals(a,b):
  return expand(a-b) == 0

# Checks to see which symbols are included in a symbolic expression
def symvar(exp):
  return sorted(list(exp.atoms(Symbol)),key = str)


# Computes the total amount of operations involved in an expresion
# or list of expressions
# @param orig_exp   An expression or list of expressions
# @return           The number of operations required to compute orig_exp
def total_ops(orig_exp):
  if type(orig_exp)==list:
    return sum((count_ops(exp_i) for exp_i in orig_exp))
  return orig_exp

# Verifies that two expressions or list of expressions is equivalent.
# If they aren't, an error was made.
# @params check       An expression to verify
# @params truth       The original expression, to make sure nothing has changed
# @params message     A debug message to print in the event that a mistake is made.
def error_check(check, truth, message = ""):
  err = ""
  if type(check)==list and type(truth)==list:
    for i in range(len(check)):
      error_check(check[i], truth[i])
  else:
    if not equals( check, truth):
      if len(message):
        err += "// "+str(message)+" made an error, input != output!\n"
      else:
        err += "// An error occurred, input != output!\n"
      err += "Error:\n\t" + str(check) + "\n\t" + str(truth) + "\n"
  return err
  

# Takes a complicated expression and generates efficient c++ code to compute it
# @params orig_exp      A large function that we need to compute efficiently
def make_efficient_code(orig_exp, sym_name = 'X', check = False, verbose = False, prefix = 'double '):
  output = ""
  code = ""
  # Count initial operations
  if verbose: 
    print "Started with"+str(total_ops(orig_exp))+"operations."
  else:
    output += "Started with "+str(total_ops(orig_exp))+" operations.\n"
  
  # Puts equations in expanded form
  if type(orig_exp) == list:
    orig_exp_expand = [0]*len(orig_exp)
    for i in range(len(orig_exp)):
      orig_exp_expand[i] = expand(orig_exp[i])
  else:
    orig_exp_expand = expand(orig_exp)

  priority_count =  mostCommonInstancesOf(sum(orig_exp_expand),symvar(sum(orig_exp_expand)))
  
  # Count post expansion operations
  if verbose: 
    print "Expanded to"+str(total_ops(orig_exp_expand))+"operations."
  else:
    output += "Expanded to "+str(total_ops(orig_exp_expand))+" operations.\n"
  
  start = timer()
  
  # Simplifying a list of expressions
  if type(orig_exp) == list:
    # Preprocess out the zeros, they don't need to be computed.
    i=0
    zeros = ''
    while i < len(orig_exp):
      if equals(orig_exp[i],0):
        zeros += prefix+sym_name[i]+' = 0.0;\n'
        del orig_exp[i]
        del sym_name[i]
      else: 
        i += 1
    # Handle a whole list of expressions
    expr = [factored(e, priority = priority_count) for e in orig_exp]
  else:
    expr = factored(orig_exp, priority = priority_count)

  if check: 
    err = error_check(expr, orig_exp, message = "Factoring algorithm")      
    if verbose: print err
    output += err

  if verbose: 
    print "After factoring there were "+str(total_ops(expr))+"operations."
  else:
    output += "After factoring there were "+str(total_ops(expr))+" operations.\n"

  # Common sub-expression extraction
  (substitution, expr) = cse_main.cse(expr)
  
  if check:  
    if type(orig_exp) == list:
      reduced = copy.deepcopy(expr)
      # Remove substitutions
      for sub in reversed(substitution):
        for i in range(len(reduced)):
  	      reduced[i] = reduced[i].subs(sub[0], sub[1])  
    else:
      expr = expr[0] # Get it out of the list, just one expression
      reduced = copy.deepcopy(expr)
      for sub in reversed(substitution):
        reduced = reduced.subs(sub[0], sub[1])
    
    if check: 
      err = error_check(expr, orig_exp, message = "CSE reduction algorithm")      
      if verbose: print err
      output += err

  # Print the results as c++ code
  for s in substitution:
    code += prefix + str(s[0])+" = "+str(s[1])+";\n"
  code += zeros[0:-1] + "\n"# We don't need the last newline char
  if type(sym_name)==list:
    for i in range(len(sym_name)):
        code += prefix + str(sym_name[i])+" = "+str(expr[i])+";\n"
  else:
    code += prefix + str(sym_name)+" = "+str(expr)+";\n"
  
  end = timer()
  sum_ops = total_ops(expr)+total_ops(substitution)
  if verbose: 
    print "Finished CSE with", sum_ops, "operations."
    print "Took",(end-start),"seconds."
    print "\n"+code
  else:
    output += "Finished CSE with "+str(sum_ops)+" operations.\n"
    output += "Took "+str(end-start)+" seconds.\n"
  return [output, code]



# Finds the most common instances of a subset of variables in an expression
# @params exp        The expression to search
# @params variables  The variables that are to be searched for
# @return            A list of (variable, count) tuples, sorted by decending count
def mostCommonInstancesOf(exp, variables):
  exp = str(exp)
  counts = {}
  for var in variables:
    counts[var] = len(list(KnuthMorrisPratt(exp, str(var))))

  count_list = list(sorted(counts, key = lambda s: counts[s], reverse = True))
  ordered = []
  for i in range(len(counts)):
    ordered.append((count_list[i], counts[count_list[i]]))

  return ordered
        


# Chooses a the variable in 'variables' that is closest to the top of the list of
# of priority
# @params variables    The variables to choose from
# @params priority     The order of preference for any variables that may come up
# @return              The highest priority variable
def pick_var(variables, priority):
  for p in priority:
    if p[0] in variables:
      return p[0]
  return variables[0]






# Factors a sum of products function that is first order with respect to all symbolic variables
# into a reduced form using products of sums whenever possible.
# @params orig_exp     A symbolic expression to be simplified
# @params depth        Used to control indenting for printing
# @params verbose      Whether to print or not
def factored(orig_exp, depth = 0, verbose = False, priority = []):
  # Prevents sympy from doing any additional factoring
  exp = orig_exp
  if verbose: tabs = '\t'*depth
  terms = []

  # Break up the added terms
  my_atoms = symvar(exp)
  if verbose: 
    print tabs,"The expression is",exp
    print tabs,my_atoms, len(my_atoms)

  # There is nothing to sort out, only one term left
  if len(my_atoms) <= 1:
    terms.append((exp, 1))
  else:
    which_atom = pick_var(my_atoms, priority)
    (c,v) = collect_terms(exp, which_atom)
    # Makes sure it doesn't factor anything extra out
    # These are the terms without variable my_atoms[0]
    if verbose: 
      print tabs, "Collecting", which_atom, "terms."
      print tabs,'Seperated terms with ',v[0], ',  (',c[0],')'
    # Factor the leftovers and recombine
    for c_subexp in range(len(c)):
      c[c_subexp] = factored(c[c_subexp], depth + 1, priority = priority)
      terms.append((v[c_subexp], c[c_subexp]))
    
  recombine = sum((terms[i][0]*terms[i][1] for i in range(len(terms))))
  # Python won't recombine complicated things with different numerical
  # coefficients unless you tell it to...
  return factor_terms(recombine, sign=False)





# Knuth-Morris-Pratt string matching
# David Eppstein, UC Irvine, 1 Mar 2002 
#from http://code.activestate.com/recipes/117214/
def KnuthMorrisPratt(text, pattern):

  '''Yields all starting positions of copies of the pattern in the text.
  Calling conventions are similar to string.find, but its arguments can be
  lists or iterators, not just strings, it returns all matches, not just
  the first one, and it does not need the whole text in memory at once.
  Whenever it yields, it will have read the text exactly up to and including
  the match that caused the yield.'''

  # allow indexing into pattern and protect against change during yield
  pattern = list(pattern)

  # build table of shift amounts
  shifts = [1] * (len(pattern) + 1)
  shift = 1
  for pos in range(len(pattern)):
      while shift <= pos and pattern[pos] != pattern[pos-shift]:
          shift += shifts[pos-shift]
      shifts[pos+1] = shift

  # do the actual search
  startPos = 0
  matchLen = 0
  for c in text:
      while matchLen == len(pattern) or \
            matchLen >= 0 and pattern[matchLen] != c:
          startPos += shifts[matchLen]
          matchLen -= shifts[matchLen]
      matchLen += 1
      if matchLen == len(pattern):
          yield startPos



